import json
import re
import requests

ACTION_TABLE_URL = 'https://raw.githubusercontent.com/salesforce/policy_sentry/master/policy_sentry/shared/data/iam-definition.json'
ACTIONS_NOT_COVERED_BY_ACCESS_ADVISOR = ['iam:PassRole', 's3:GetObject', 's3:PutObject']

action_map = {a['prefix']: a for a in requests.get(ACTION_TABLE_URL).json()}


class PolicyAnalyzer:
    @staticmethod
    def convert_to_list(list_or_single_object):
        if isinstance(list_or_single_object, list):
            return list_or_single_object
        return [list_or_single_object]

    @staticmethod
    def _get_policy_actions(policy_document: dict):
        policy_statements = PolicyAnalyzer.convert_to_list(policy_document['Statement'])
        actions_list = []
        for statement in policy_statements:
            if statement['Effect'] == 'Allow':
                actions_list.extend(PolicyAnalyzer.convert_to_list(statement['Action']))
        return actions_list

    @staticmethod
    def is_policy_unused(policy_document: dict, services_last_accessed: list) -> bool:
        statements_str = json.dumps(policy_document['Statement'])
        if '"Effect": "Deny"' in statements_str or '"NotAction":' in statements_str:
            # If statement contains a "Deny" effect - Access Advisor won't detect that action because it is a restriction
            # If statement contains a "NotAction" effect - Access Advisor won't detect usage of this policy correctly
            return False

        policy_actions = PolicyAnalyzer._get_policy_actions(policy_document)
        if len([action for action in policy_actions if
                len(list(filter(re.compile(action.replace('*', '.*')).match, ACTIONS_NOT_COVERED_BY_ACCESS_ADVISOR))) > 0]) > 0:
            return False

        services_accessed_through_policy = list(set(map(lambda action: action.split(':')[0], policy_actions)))
        return len(
            [service for service in services_accessed_through_policy if
             len(list(filter(re.compile(service.replace('*', '.*')).match, services_last_accessed))) > 0
             ]) == 0

    @staticmethod
    def policy_is_write_access(policy_document):
        actions = PolicyAnalyzer._get_policy_actions(policy_document)
        for action in actions:
            if action == '*' or '*' in action.split(':'):
                return True
            [action_service, action_name] = action.split(':')
            try:
                action_regex = action_name.replace('*', '.*')
                action_objs = list(filter(lambda privilege_obj: re.match(action_regex, privilege_obj['privilege']),
                                          action_map[action_service]['privileges']))
            except StopIteration:
                action_objs = []

            for action_obj in action_objs:
                if action_obj['access_level'] in ['Write', 'Delete', 'Permissions management']:
                    return True
        return False



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer, Principal
from airiam.terraform.entity_terraformers.IAMPolicyDocumentTransformer import IAMPolicyDocumentTransformer


class IAMInlinePolicyTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict, principal_name: str, principal: Principal):
        policy_name = BaseEntityTransformer.safe_name_converter(entity_json['PolicyName'])
        self.principal_name = principal_name
        self._safe_user_name = BaseEntityTransformer.safe_name_converter(self.principal_name)
        self._principal = principal.value
        super().__init__(f"aws_iam_{principal.value}_policy", policy_name, entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        policy_document_hcl = IAMPolicyDocumentTransformer(entity_json['PolicyDocument'], f"{self._safe_name}_document", self.principal_name)
        return f"""resource "aws_iam_{self._principal}_policy" "{self.principal_name}_{self._safe_name}" {{
  name   = "{entity_json['PolicyName']}"
  policy = {policy_document_hcl.identifier()}.json
  {self._principal}   = aws_iam_{self._principal}.{self._safe_user_name}.name
}}

{policy_document_hcl.code()}
"""

    def entities_to_import(self) -> list:
        return [{"identifier": f"aws_iam_{self._principal}_policy.{self.principal_name}_{self._safe_name}", "entity": f"{self.principal_name}:{self._entity_name}"}]



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer

import json


class IAMGroupMembershipsTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict):
        super().__init__('aws_iam_group_membership', f"{entity_json['GroupName']}_membership", entity_json)
        self._users = entity_json['Users']

    def _generate_hcl2_code(self, entity_json) -> str:
        return f"""resource "aws_iam_group_membership" "{self._safe_name}" {{
  name = "{self._safe_name}"
  group = {entity_json['GroupHcl']}.name
  users = {json.dumps(entity_json['Users'])}
}}

"""

    def entities_to_import(self) -> list:
        entities_to_import = []
        for i in range(0, len(self._users)):
            entities_to_import.append({"identifier": f"{self.identifier()}[i]", })
        return entities_to_import



import argparse
import logging
import sys

from airiam.Reporter import Reporter, OutputFormat
from airiam.find_unused.find_unused import find_unused
from airiam.recommend_groups.recommend_groups import recommend_groups
from airiam.terraform.TerraformTransformer import TerraformTransformer


def configure_logger(logging_level=logging.INFO):
    logging.basicConfig(level=logging_level)
    # define a Handler which writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    # set a format which is simpler for console use
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    # tell the handler to use this format
    console.setFormatter(formatter)
    return logging


def run():
    logger = configure_logger()

    Reporter.print_prelude()
    args = parse_args(sys.argv[1:])

    runtime_results = find_unused(logger, args.profile, args.no_cache, args.last_used_threshold, args.command)

    if args.command == 'find_unused':
        Reporter.report_unused(runtime_results)
        exit()

    if args.command == 'recommend_groups' or args.command == 'terraform' and not args.without_groups:
        report_with_recommendations = recommend_groups(logger, runtime_results, args.last_used_threshold)
        if args.command == 'recommend_groups':
            Reporter.report_groupings(report_with_recommendations)
            exit()

    if args.command == 'terraform':
        entities_terraformed, result_dir = TerraformTransformer(logger, args.profile, args.directory)\
            .transform(runtime_results, args.without_unused, args.without_groups, args.without_import)
        Reporter.report_terraform(entities_terraformed, result_dir)


def parse_args(args):
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--version', help='Get AirIAM\'s version', action='store_true')

    sub_parsers = parser.add_subparsers(title='commands', dest='command')
    find_unused_parser = sub_parsers.add_parser('find_unused', help='Scan your runtime IAM for unused entities',
                                                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    find_unused_parser.add_argument('-p', '--profile', help='AWS profile to be used', type=str, default=None)
    find_unused_parser.add_argument('-l', '--last-used-threshold', help='"Last Used" threshold, in days, for an entity to be considered unused',
                                    type=int, default=90)
    find_unused_parser.add_argument('--no-cache', help='Generate a fresh set of data from AWS IAM API calls', action='store_true')
    find_unused_parser.add_argument('-o', '--output', help='Output format', type=OutputFormat,
                                    choices=[output.name for output in OutputFormat], default=OutputFormat.cli)

    recommend_groups_parser = sub_parsers.add_parser('recommend_groups', help='Recommend IAM groups according to IAM users and their in-use privileges',
                                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    recommend_groups_parser.add_argument('-p', '--profile', help='AWS profile to be used', type=str, default=None)
    recommend_groups_parser.add_argument('-o', '--output', help='Output format', type=OutputFormat,
                                         choices=[output.name for output in OutputFormat], default=OutputFormat.cli)
    recommend_groups_parser.add_argument('-l', '--last-used-threshold', type=int, default=90,
                                         help='"Last Used" threshold, in days, for an entity to be considered unused')
    recommend_groups_parser.add_argument('--no-cache', help='Generate a fresh set of data from AWS IAM API calls', action='store_true')

    tf_parser = sub_parsers.add_parser('terraform', help='Terraformize your runtime AWS IAM configurations',
                                       formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    tf_parser.add_argument('-p', '--profile', help='AWS profile to be used', type=str)
    tf_parser.add_argument('-d', '--directory', help='Path where the output terraform code and state will be stored', type=str, default='results')
    tf_parser.add_argument('--without-unused', help='Create terraform code without unused entities', action='store_true')
    tf_parser.add_argument('--without-groups', help='Create terraform code without recommendation for user groups', action='store_true')
    tf_parser.add_argument('-l', '--last-used-threshold', help='"Last Used" threshold, in days, for an entity to be considered unused', type=int,
                           default=90)
    tf_parser.add_argument('--no-cache', help='Generate a fresh set of data from AWS IAM API calls', action='store_true')
    tf_parser.add_argument('--without-import', help='Import the resulting entities to terraform\'s state file. Note - this might take a long time',
                           action='store_true')
    result = parser.parse_args(args)
    if result.version:
        Reporter.print_version()
        exit(0)
    if not result.command:
        parser.print_help(sys.stderr)
        sys.exit(1)
    return result



import time
from enum import Enum

from colorama import init
from termcolor import colored

from airiam.banner import banner
from airiam.models import RuntimeReport
from airiam.version import version

init(autoreset=True)


class OutputFormat(Enum):
    cli = 'CLI'


class Reporter:
    @staticmethod
    def report_unused(runtime_results: RuntimeReport) -> None:
        print(f'Identifying unused IAM entities in the account...\n')
        time.sleep(2)
        unused = runtime_results.get_unused()
        unused_users = unused['Users']
        if len(unused_users) > 0:
            print(colored(f'The following {len(unused_users)} users were found to be unused:', 'yellow', attrs=['bold']))
            for user in unused_users:
                print(colored('Unused: ', 'red', attrs=['bold']) + f'{user["UserName"]}: last used {user["LastUsed"]} days ago')
            time.sleep(5)
        else:
            print(colored('No unused users were found in the account! Hurray!', color='green'))

        print()
        unused_access_keys = unused['UnusedActiveAccessKeys']
        if len(unused_access_keys) > 0:
            print(colored(f'The following {len(unused_access_keys)} active access keys aren\'t being used:', 'yellow', attrs=['bold']))
            for access_key_obj in unused_access_keys:
                print(colored('Unused: ', 'red', attrs=['bold'])
                      + f'{access_key_obj["User"]} used access key #{access_key_obj["AccessKey"]} {access_key_obj["DaysSinceLastUse"]} days ago')
            time.sleep(5)
        else:
            print(colored('No unused access keys were found in the account! Hurray!', color='green'))

        print()
        unused_console_login_profiles = unused['UnusedConsoleLoginProfiles']
        if len(unused_console_login_profiles) > 0:
            print(colored(f'The following {len(unused_console_login_profiles)} users have password access they aren\'t using:', 'yellow', attrs=['bold']))
            for console_login_profile in unused_console_login_profiles:
                has_mfa = console_login_profile['MFAEnabled']
                if has_mfa:
                    print(colored(console_login_profile['User'], 'yellow', attrs=['bold'])
                          + ' has password access to the AWS console (with MFA) but hasn\'t used it in the last '
                          + f'{console_login_profile["DaysSinceLastUse"]} days')
                else:
                    print(colored(console_login_profile['User'], 'red', attrs=['bold'])
                          + f' has password access to the AWS console ' + colored('WITHOUT MFA', 'red')
                          + f' but hasn\'t used it in the last {console_login_profile["DaysSinceLastUse"]} days')
            time.sleep(5)
        else:
            print(colored('No unused Console Login Profiles were found in the account! Hurray!', color='green'))

        print()
        unused_roles = unused['Roles']
        if len(unused_roles) > 0:
            print(colored(f'The following {len(unused_roles)} roles are unused:', 'yellow', attrs=['bold']))
            for role in unused_roles:
                print(colored('Unused: ', 'red', attrs=['bold']) + f"{role['RoleName']}: last used {role['LastUsed']} days ago")
            time.sleep(5)
        else:
            print(colored('No unused roles were found in the account! Hurray!', color='green'))

        print()
        unused_groups = unused['Groups']
        if len(unused_groups) > 0:
            print(colored(f'The following {len(unused_groups)} groups are redundant:', 'yellow', attrs=['bold']))
            for group in unused_groups:
                if len(group['AttachedManagedPolicies'] + group['GroupPolicyList']) == 0:
                    msg = ' has no policies attached to it'
                else:
                    msg = ' has no members'
                print(colored(group['GroupName'], 'yellow', attrs=['bold']) + msg)
            time.sleep(5)
        else:
            print(colored('No redundant groups were found in the account! Hurray!', color='green'))

        print()
        unused_policies = unused['Policies']
        if len(unused_policies) > 0:
            print(colored(f'The following {len(unused_policies)} policies are redundant:', 'yellow', attrs=['bold']))
            for policy in unused_policies:
                print(colored(policy['PolicyName'], 'yellow', attrs=['bold']) + f' is not attached to any user, group or role')
            time.sleep(5)
        else:
            print(colored('No unattached policies were found in the account! Hurray!', color='green'))

        print()
        unused_policy_attachments = unused['PolicyAttachments']
        if len(unused_policy_attachments) > 0:
            print(colored(f'The following {len(unused_policy_attachments)} policy attachments are unused:', 'yellow', attrs=['bold']))
            for policy_attachment in unused_policy_attachments:
                principal = policy_attachment.get('Role') or policy_attachment.get('User') or policy_attachment.get('Group')
                print(colored('Policy attached but not used: ', 'yellow', attrs=['bold']) + colored(principal, 'grey', attrs=['bold']) +
                      f' is not using the privileges given by {colored(policy_attachment["PolicyArn"], "red", attrs=["bold"])}')
            time.sleep(5)
        else:
            print(colored('No unused policy attachments were found in the account! Hurray!', color='green'))

        print()

        print('If you prefer to to change the current runtime and not move to IaC but the number of entities above is intimidating - consider using '
              'our playbooks, available at: ')
        print('https://www.bridgecrew.io/')

    @staticmethod
    def print_prelude():
        print(colored(banner, 'yellow'))
        print()
        print(colored('AirIAM - Least privilege AWS IAM Terraformer', attrs=['bold']))
        print()
        print(f'To continuously scan configurations, try the {colored("Bridgecrew", "magenta", attrs=["bold"])} free community plan.')
        print('https://www.bridgecrew.io')
        print()

    @staticmethod
    def report_groupings(report_with_recommendations: RuntimeReport):
        simple_user_clusters = report_with_recommendations.get_user_groups()
        admins = simple_user_clusters['Admins']
        read_only = simple_user_clusters['ReadOnly']
        powerusers = simple_user_clusters['Powerusers']
        print(colored(f'\nThe following {len(admins["Users"])} users require admin access to the account:', 'yellow', attrs=['bold']))
        for user in admins['Users']:
            print(colored('Admin: ', 'red', attrs=['bold']) + user)
        print()
        if len(powerusers['Users']) > 0:
            print(f'The following {len(powerusers["Users"])} users require ' +
                  colored('partial write', 'yellow', attrs=['bold']) +
                  ' access to the account:')
            for user in powerusers['Users']:
                print(colored('Poweruser: ', 'yellow', attrs=['bold']) + user)
        print()
        if len(read_only['Users']) > 0:
            print(colored(f'The following {len(read_only["Users"])} users require ReadOnly access to the account:', 'grey', attrs=['bold']))
            for user in read_only['Users']:
                print(colored('ReadOnly: ', 'green') + user)


    @staticmethod
    def report_terraform(terraformed_entities: dict, result_dir: str):
        print(colored('Successfully ', 'green', attrs=['bold']) + 'migrated your current IAM setup to terraform!')
        print(f'Migrated {len(terraformed_entities["Users"])} users, {len(terraformed_entities["Groups"])} groups, {len(terraformed_entities["Roles"])} '
              f'roles and {len(terraformed_entities["Policies"])} policies, as well as all connections between them, to terraform.')
        print(f'Your terraform files can now be found at the directory you specified: ' + colored(result_dir, attrs=['underline']))

    @classmethod
    def print_version(cls):
        print(f"AirIAM version {version}")



from python_terraform import *
import copy

from airiam.models import RuntimeReport
from airiam.terraform.entity_terraformers.AWSProviderTransformer import AWSProviderTransformer
from airiam.terraform.entity_terraformers.IAMGroupTransformer import IAMGroupTransformer
from airiam.terraform.entity_terraformers.IAMPolicyTransformer import IAMPolicyTransformer
from airiam.terraform.entity_terraformers.IAMRoleTransformer import IAMRoleTransformer
from airiam.terraform.entity_terraformers.IAMUserGroupMembershipTransformer import IAMUserGroupMembershipTransformer
from airiam.terraform.entity_terraformers.IAMUserTransformer import IAMUserTransformer

current_dir = os.path.abspath(os.path.dirname(__file__))
boilerplate_files = ["admins.tf", "developers.tf", "power_users.tf"]
ERASE_LINE = '\x1b[2K'


class TerraformTransformer:
    def __init__(self, logger, profile=None, result_dir='results'):
        self.logger = logger
        self.profile = profile
        self._result_dir = result_dir
        if not os.path.exists(self._result_dir):
            os.mkdir(self._result_dir)

    def transform(self, results: RuntimeReport, without_unused: bool, without_groups: bool, without_import: bool) -> (dict, str):
        try:
            if not without_groups:
                # todo: implement!
                self.logger.warn('Migrating to the recommended groups isn\'t supported yet. Issue exists - '
                                 'https://github.com/bridgecrewio/AirIAM/issues/20')
                self.logger.warn('Will use the existing groups for terraform migration until it is implemented')
            entities_to_transform = self._list_entities_to_transform(results, without_unused, without_groups)
            entities_to_import = self.write_terraform_code(entities_to_transform)
            tf = Terraform(working_dir=self._result_dir)
            tf.init(backend=False)
            tf.fmt()
            if not without_import:
                num_of_entities_to_import = len(entities_to_import)
                print(f"Importing {num_of_entities_to_import} entities")
                i = 1
                for entity_to_import in entities_to_import:
                    msg = f"#{i} of {num_of_entities_to_import}: Importing {entity_to_import['entity']} to {entity_to_import['identifier']}"
                    print(ERASE_LINE + f"\r{msg}", end="")
                    return_code, stdout, stderr = tf.import_cmd(entity_to_import['identifier'], entity_to_import['entity'])
                    if return_code != 0 and 'Resource already managed by Terraform' not in stderr:
                        self.logger.error(f"Error: {stderr}")
                    i += 1
                print("Imported all existing entities to state")

            return entities_to_transform, self._result_dir
        except Exception as e:
            self.logger.error(e)
            raise e

    def _list_entities_to_transform(self, report: RuntimeReport, without_unused: bool, without_consolidated_groups: bool) -> dict:
        iam_raw_data = report.get_raw_data()
        raw_entities_to_transform = {
            'Users': copy.copy(iam_raw_data['AccountUsers']),
            'Roles': copy.copy(iam_raw_data['AccountRoles']),
            'Policies': copy.copy(iam_raw_data['AccountPolicies']),
            'Groups': copy.copy(iam_raw_data['AccountGroups'])
        }
        if without_unused:
            unused = report.get_unused()
            unused_entities = unused['Users'] + unused['Roles'] + unused['Policies'] + unused['Groups'] + unused['PolicyAttachments']
            self.logger.warn(f'Filtering out {len(unused_entities)} entities from terraform. These entities will have to be handled manually')
            for user in unused['Users']:
                raw_entities_to_transform['Users'].remove(user)
            for role in unused['Roles']:
                raw_entities_to_transform['Roles'].remove(role)
            for group in unused['Groups']:
                raw_entities_to_transform['Groups'].remove(group)
            for policy in unused['Policies']:
                raw_entities_to_transform['Policies'].remove(policy)
            for policy_attachment_obj in unused['PolicyAttachments']:
                if 'Role' in policy_attachment_obj:
                    TerraformTransformer.remove_from_transformation(policy_attachment_obj, raw_entities_to_transform, 'Role')
                elif 'User' in policy_attachment_obj:
                    TerraformTransformer.remove_from_transformation(policy_attachment_obj, raw_entities_to_transform, 'User')
                elif 'Group' in policy_attachment_obj:
                    TerraformTransformer.remove_from_transformation(policy_attachment_obj, raw_entities_to_transform, 'Group')

        # todo: iterate over users fix group attachments
        # todo: if without_unused, delete older groups
        # if not without_consolidated_groups:

        return raw_entities_to_transform

    @staticmethod
    def remove_from_transformation(policy_attachment_obj: dict, entity_dict: dict, principal_type: str):
        """
        This method removes the policy attachment from the entity list. It relies heavily on the reuse of the structures by AWS and AirIAM:
        The entity list will always be named the <principal_type>s, e.g. Users/Roles/Groups
        The key for the inline policy list will always be named <principal_type>PolicyList, e.g. RolePolicyList / UserPolicyList
        The key for the principal_id will be the principal_type itself, i.e. Group / User / Role
        :param policy_attachment_obj: A dict which contains one of the following keys: User, Role, Group
        :param entity_dict:           A dict which holds all the entities as lists with the relevant key being the principal_type
        :param principal_type:        One of: User / Role / Group
        """
        policy_id = policy_attachment_obj['PolicyArn']
        principal = next(p for p in entity_dict[f'{principal_type}s'] if p[f'{principal_type}Name'] == policy_attachment_obj[principal_type])
        if policy_id.startswith('arn:aws'):
            policy_attachment = next(policy for policy in principal['AttachedManagedPolicies']
                                     if policy['PolicyArn'] == policy_attachment_obj['PolicyArn'])
            principal['AttachedManagedPolicies'].remove(policy_attachment)
        else:
            policy_attachment = next(policy for policy in principal[f'{principal_type}PolicyList']
                                     if policy['PolicyName'] == (policy_attachment_obj.get('PolicyName') or policy_attachment_obj.get('PolicyArn')))
            principal[f'{principal_type}PolicyList'].remove(policy_attachment)

    def write_terraform_code(self, iam_entities: dict) -> list:
        entities_to_import = []
        with open(f"{self._result_dir}/main.tf", 'w') as main_file:
            main_code = AWSProviderTransformer({'region': 'us-east-1', 'profile': self.profile}).code()
            main_file.write(main_code)

        policies_identifiers = {}

        with open(f"{self._result_dir}/policies.tf", 'w') as policies_file:
            policy_code = ""
            for policy in iam_entities['Policies']:
                if 'iam::aws:' in policy['Arn']:
                    # Don't create AWS managed policies
                    continue
                policy_transformer = IAMPolicyTransformer(policy)
                policy_code += policy_transformer.code()
                policies_identifiers[policy['Arn']] = policy_transformer.identifier()
                entities_to_import += policy_transformer.entities_to_import()
            policies_file.write(policy_code)

        with open(f"{self._result_dir}/groups.tf", 'w') as groups_file:
            groups_code = ""
            groups_identifiers = {}
            for group in iam_entities['Groups']:
                group_transformer = IAMGroupTransformer(group)
                groups_code += group_transformer.code()
                groups_identifiers[group['GroupName']] = group_transformer.identifier()
                entities_to_import += group_transformer.entities_to_import()
            groups_file.write(groups_code)

        user_group_memberships = {}
        with open(f"{self._result_dir}/users.tf", 'w') as users_file:
            user_code = ""
            for user in iam_entities['Users']:
                transformer = IAMUserTransformer(user)
                user_code += transformer.code()
                for group in user['GroupList']:
                    if group not in user_group_memberships:
                        user_group_memberships[group] = []
                    user_group_memberships[group].append(user['UserName'])
                membership_transformer = IAMUserGroupMembershipTransformer({"UserName": user['UserName'], "Groups": user['GroupList']},
                                                                           transformer.identifier())
                user_code += membership_transformer.code()
                entities_to_import += transformer.entities_to_import() + membership_transformer.entities_to_import()
            users_file.write(user_code)

        with open(f"{self._result_dir}/roles.tf", 'w') as roles_file:
            roles_code = ""
            for role in iam_entities['Roles']:
                transformer = IAMRoleTransformer(role)
                roles_code += transformer.code()
                entities_to_import += transformer.entities_to_import()
            roles_file.write(roles_code)

        return entities_to_import



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer, Principal


class IAMManagedPolicyAttachmentTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict, principal_name: str, principal_type: Principal):
        self._user_name = principal_name
        self._safe_user_name = BaseEntityTransformer.safe_name_converter(principal_name)
        self._principal = principal_type.value
        self._policy_arn = entity_json['PolicyArn']
        attachment_name = f"{self._safe_user_name}_{BaseEntityTransformer.safe_name_converter(entity_json['PolicyName'])}"
        if self.is_aws_managed(entity_json['PolicyArn']):
            attachment_name = f"{self._safe_user_name}_{BaseEntityTransformer.safe_name_converter(entity_json['PolicyName'])}_managed"
        super().__init__(f"aws_iam_{self._principal}_policy_attachment", attachment_name, entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        policy_arn = f'"{entity_json["PolicyArn"]}"'
        if not self.is_aws_managed(entity_json['PolicyArn']):
            policy_arn = f"aws_iam_policy.{BaseEntityTransformer.safe_name_converter(entity_json['PolicyName'])}.arn"

        return f"""resource "aws_iam_{self._principal}_policy_attachment" "{self._safe_name}" {{
  policy_arn = {policy_arn}
  {self._principal} = aws_iam_{self._principal}.{self._safe_user_name}.name
}}
"""

    def entities_to_import(self) -> list:
        return [{"identifier": self.identifier(), "entity": f"{self._user_name}/{self._policy_arn}"}]

    @staticmethod
    def is_aws_managed(policy_arn):
        return 'iam::aws:' in policy_arn



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer


class InstanceProfileTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict, role_identifier: str):
        self.raw_name = entity_json['Arn'].split('/')[-1]
        self._role_identifier = role_identifier
        super().__init__('aws_iam_instance_profile', BaseEntityTransformer.safe_name_converter(self.raw_name), entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        tags = self.transform_tags(entity_json)

        return f"""resource "{self._entity_type}" "{self._safe_name}" {{
  name = "{self.raw_name}"
  path = "{entity_json['Path']}"
  role = {self._role_identifier}.name
}}

"""



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer


class AWSProviderTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict):
        super().__init__('aws_provider', 'default_provider', entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        if 'profile' in entity_json and entity_json['profile'] is not None:
            provider_code = f"""provider "aws" {{
  region  = "{entity_json['region']}"
  profile = "{entity_json['profile']}"
}}

"""
        else:
            provider_code = f"""provider "aws" {{
  region = "{entity_json['region']}"
}}

"""
        return provider_code

    def entities_to_import(self) -> list:
        return []



SORT_KEY_BY_ENTITY_TYPE = {
    'AccountGroups': 'GroupName',
    'AccountPolicies': 'Arn',
    'AccountRoles': 'RoleName',
    'AccountUsers': 'UserName'
}


class RuntimeReport:
    def __init__(self, account_id: str, identity_arn: str, raw_results: dict):
        self.account_id = account_id
        self.identity_arn = identity_arn
        self._raw_results = raw_results
        self._unused_users = None
        self._unused_roles = None
        self._unattached_policies = None
        self._redundant_groups = None
        self._unused_active_access_keys = None
        self._unused_console_login_profiles = None
        self._unused_policy_attachments = None
        self._user_group_recommendation = None

    def get_raw_data(self) -> dict:
        return self._raw_results

    def get_unused(self) -> dict:
        return {
            'Users': self._unused_users,
            'UnusedActiveAccessKeys': self._unused_active_access_keys,
            'UnusedConsoleLoginProfiles': self._unused_console_login_profiles,
            'Roles': self._unused_roles,
            'Policies': self._unattached_policies,
            'Groups': self._redundant_groups,
            'PolicyAttachments': self._unused_policy_attachments
        }

    def get_user_groups(self) -> dict:
        return self._user_group_recommendation

    def set_unused(self, unused_users, unused_roles, unused_active_access_keys, unused_console_login_profiles, unattached_policies, redundant_groups,
                   unused_policy_attachments):
        self._unused_users = sorted(unused_users, key=lambda user: SORT_KEY_BY_ENTITY_TYPE['AccountUsers'])
        self._unused_roles = sorted(unused_roles, key=lambda role: SORT_KEY_BY_ENTITY_TYPE['AccountRoles'])
        self._unused_active_access_keys = sorted(unused_active_access_keys, key=lambda access_key: access_key['User'])
        self._unused_console_login_profiles = sorted(unused_console_login_profiles, key=lambda access_key: access_key['User'])
        self._unattached_policies = sorted(unattached_policies, key=lambda policy: SORT_KEY_BY_ENTITY_TYPE['AccountPolicies'])
        self._redundant_groups = sorted(redundant_groups, key=lambda group: SORT_KEY_BY_ENTITY_TYPE['AccountGroups'])
        self._unused_policy_attachments = sorted(unused_policy_attachments, key=RuntimeReport.policy_attachment_sorter)

    @staticmethod
    def policy_attachment_sorter(policy_attachment):
        return policy_attachment.get('Role') or policy_attachment.get('User') or policy_attachment.get('Group')

    def set_reorg(self, user_group_recommendation):
        self._user_group_recommendation = user_group_recommendation
